# Created on 2018/12
# Author: Kaituo XU

import math

import torch


def overlap_and_add(signal, frame_step):
    """Reconstructs a signal from a framed representation.

    Adds potentially overlapping frames of a signal with shape
    `[..., frames, frame_length]`, offsetting subsequent frames by `frame_step`.
    The resulting tensor has shape `[..., output_size]` where

        output_size = (frames - 1) * frame_step + frame_length

    Args:
        signal: A [..., frames, frame_length] Tensor. All dimensions may be unknown, and rank must be at least 2.
        frame_step: An integer denoting overlap offsets. Must be less than or equal to frame_length.

    Returns:
        A Tensor with shape [..., output_size] containing the overlap-added frames of signal's inner-most two dimensions.
        output_size = (frames - 1) * frame_step + frame_length

    Based on https://github.com/tensorflow/tensorflow/blob/r1.12/tensorflow/contrib/signal/python/ops/reconstruction_ops.py
    """
    outer_dimensions = signal.size()[:-2]
    frames, frame_length = signal.size()[-2:]

    subframe_length = math.gcd(frame_length, frame_step)  # gcd=Greatest Common Divisor
    subframe_step = frame_step // subframe_length
    subframes_per_frame = frame_length // subframe_length
    output_size = frame_step * (frames - 1) + frame_length
    output_subframes = output_size // subframe_length

    subframe_signal = signal.view(*outer_dimensions, -1, subframe_length)

    frame = torch.arange(0, output_subframes).unfold(0, subframes_per_frame, subframe_step)
    frame = signal.new_tensor(frame).long()  # signal may in GPU or CPU
    frame = frame.contiguous().view(-1)

    result = signal.new_zeros(*outer_dimensions, output_subframes, subframe_length)
    result.index_add_(-2, frame, subframe_signal)
    result = result.view(*outer_dimensions, -1)
    return result


def remove_pad(inputs, inputs_lengths):
    """
    Args:
        inputs: torch.Tensor, [B, C, T] or [B, T], B is batch size
        inputs_lengths: torch.Tensor, [B]
    Returns:
        results: a list containing B items, each item is [C, T], T varies
    """
    results = []
    dim = inputs.dim()
    if dim == 3:
        C = inputs.size(1)
    for input, length in zip(inputs, inputs_lengths):
        if dim == 3: # [B, C, T]
            results.append(input[:,:length].view(C, -1).cpu().numpy())
        elif dim == 2:  # [B, T]
            results.append(input[:length].view(-1).cpu().numpy())
    return results


if __name__ == '__main__':
    torch.manual_seed(123)
    M, C, K, N = 2, 2, 3, 4
    frame_step = 2
    signal = torch.randint(5, (M, C, K, N))
    result = overlap_and_add(signal, frame_step)
    print(signal)
    print(result)


"""resnext in pytorch



[1] Saining Xie, Ross Girshick, Piotr Dollár, Zhuowen Tu, Kaiming He.

    Aggregated Residual Transformations for Deep Neural Networks
    https://arxiv.org/abs/1611.05431
"""

import math
import torch
import torch.nn as nn
import torch.nn.functional as F

#only implements ResNext bottleneck c


#"""This strategy exposes a new dimension, which we call “cardinality”
#(the size of the set of transformations), as an essential factor
#in addition to the dimensions of depth and width."""
CARDINALITY = 32
DEPTH = 4
BASEWIDTH = 64

#"""The grouped convolutional layer in Fig. 3(c) performs 32 groups
#of convolutions whose input and output channels are 4-dimensional.
#The grouped convolutional layer concatenates them as the outputs
#of the layer."""

class ResNextBottleNeckC(nn.Module):

    def __init__(self, in_channels, out_channels, stride):
        super().__init__()

        C = CARDINALITY #How many groups a feature map was splitted into

        #"""We note that the input/output width of the template is fixed as
        #256-d (Fig. 3), We note that the input/output width of the template
        #is fixed as 256-d (Fig. 3), and all widths are dou- bled each time
        #when the feature map is subsampled (see Table 1)."""
        D = int(DEPTH * out_channels / BASEWIDTH) #number of channels per group
        self.split_transforms = nn.Sequential(
            nn.Conv1d(in_channels, C * D, kernel_size=1, groups=C, bias=False),
            nn.BatchNorm1d(C * D),
            nn.ReLU(inplace=True),
            nn.Conv1d(C * D, C * D, kernel_size=3, stride=stride, groups=C, padding=1, bias=False),
            nn.BatchNorm1d(C * D),
            nn.ReLU(inplace=True),
            nn.Conv1d(C * D, out_channels * 4, kernel_size=1, bias=False),
            nn.BatchNorm1d(out_channels * 4),
        )

        self.shortcut = nn.Sequential()

        if stride != 1 or in_channels != out_channels * 4:
            self.shortcut = nn.Sequential(
                nn.Conv1d(in_channels, out_channels * 4, stride=stride, kernel_size=1, bias=False),
                nn.BatchNorm1d(out_channels * 4)
            )

    def forward(self, x):
        return F.relu(self.split_transforms(x) + self.shortcut(x))

class ResNext(nn.Module):

    def __init__(self, block, num_blocks, class_names=2):
        super().__init__()
        self.in_channels = 64

        self.conv1 = nn.Sequential(
            nn.Conv1d(1, 64, 3, stride=1, padding=1, bias=False),
            nn.BatchNorm1d(64),
            nn.ReLU(inplace=True)
        )

        self.conv2 = self._make_layer(block, num_blocks[0], 64, 1)
        self.conv3 = self._make_layer(block, num_blocks[1], 128, 2)
        self.conv4 = self._make_layer(block, num_blocks[2], 256, 2)
        self.conv5 = self._make_layer(block, num_blocks[3], 512, 2)
        self.avg = nn.AdaptiveAvgPool1d(1)
        self.fc = nn.Linear(512 * 4, 2)

    def forward(self, x):
        b_size = x.shape[0]
        x = self.conv1(x)
        x = self.conv2(x)
        x = self.conv3(x)
        x = self.conv4(x)
        x = self.conv5(x)
        x = self.avg(x)
        x = x.view(x.size(0), -1)
        # print(x.size())
        x = self.fc(x)
        x =x.reshape(b_size, -1, 2)
        # print(x.size())
        return x

    def _make_layer(self, block, num_block, out_channels, stride):
        """Building resnext block
        Args:
            block: block type(default resnext bottleneck c)
            num_block: number of blocks per layer
            out_channels: output channels per block
            stride: block stride

        Returns:
            a resnext layer
        """
        strides = [stride] + [1] * (num_block - 1)
        layers = []
        for stride in strides:
            layers.append(block(self.in_channels, out_channels, stride))
            self.in_channels = out_channels * 4

        return nn.Sequential(*layers)

def resnext50():
    """ return a resnext50(c32x4d) network
    """
    return ResNext(ResNextBottleNeckC, [3, 4, 6, 3])

def resnext101():
    """ return a resnext101(c32x4d) network
    """
    return ResNext(ResNextBottleNeckC, [3, 4, 23, 3])

def resnext152():
    """ return a resnext101(c32x4d) network
    """
    return ResNext(ResNextBottleNeckC, [3, 4, 36, 3])






import torch





def negative_si_snr(x, s, eps=1e-13):
    """
    Args:
        x: Enhanced fo shape [B, T]
        s: Reference of shape [B, T]
    Returns:
        si_snr: [B]
    """

    def l2norm(mat, keep_dim=False):
        return torch.norm(mat, dim=-1, keepdim=keep_dim)

    # Zero-mean x and s
    x_mean = torch.mean(x, dim=-1, keepdim=True)
    s_mean = torch.mean(s, dim=-1, keepdim=True)
    x_zm = x - x_mean
    s_zm = s - s_mean

    # Scaling factor for the target
    t = torch.sum(x_zm * s_zm, dim=-1, keepdim=True) * s_zm / (l2norm(s_zm, keep_dim=True) ** 2 + eps)

    # SI-SNR computation
    si_snr = -torch.mean(20 * torch.log10(eps + l2norm(t) / (l2norm(x_zm - t) + eps)))

    return si_snr


from torchsense.transforms.augmentations.normalize import Normalize
from torchsense.transforms.augmentations.to_tensor import ToTensor
from torchsense.transforms.augmentations.interpolate import Interpolate
from torchsense.transforms.augmentations.griffinlim import GriffinLim
from torchsense.transforms.augmentations.premodule import PreModule
from torchsense.transforms.augmentations.utils import has_batch_dimension, add_batch_dimension, remove_batch_dimension
from torchsense.transforms.augmentations.addnoise import AddNoise



import numpy as np
import torch
import numpy.random as random


class Normalize(torch.nn.Module):
    def __init__(self, new_min=0, new_max=1):
        """
        :param min_snr: Minimum signal-to-noise ratio
        :param max_snr: Maximum signal-to-noise ratio
        """
        super().__init__()
        self.min_snr = new_min
        self.max_snr = new_max

    def forward(self, x):

        x = (x - x.min()) / (x.max() - x.min())  # 先将x归一化到0~1
        x = x * (self.min_snr - self.max_snr) + self.max_snr  # 然后将x的范围调整到new_min~new_max
        return x



import torch
import torch.nn as nn
from torchinfo import summary


class GlobalLayerNorm(nn.Module):
    '''
       Calculate Global Layer Normalization
       dim: (int or list or torch.Size) –
            input shape from an expected input of size
       eps: a value added to the denominator for numerical stability.
       elementwise_affine: a boolean value that when set to True, 
           this module has learnable per-element affine parameters 
           initialized to ones (for weights) and zeros (for biases).
    '''

    def __init__(self, dim, eps=1e-05, elementwise_affine=True):
        super(GlobalLayerNorm, self).__init__()
        self.dim = dim
        self.eps = eps
        self.elementwise_affine = elementwise_affine

        if self.elementwise_affine:
            self.weight = nn.Parameter(torch.ones(self.dim, 1))
            self.bias = nn.Parameter(torch.zeros(self.dim, 1))
        else:
            self.register_parameter('weight', None)
            self.register_parameter('bias', None)

    def forward(self, x):
        # x = N x C x L
        # N x 1 x 1
        # cln: mean,var N x 1 x L
        # gln: mean,var N x 1 x 1
        if x.dim() != 3:
            raise RuntimeError("{} accept 3D tensor as input".format(
                self.__name__))

        mean = torch.mean(x, (1, 2), keepdim=True)
        var = torch.mean((x - mean) ** 2, (1, 2), keepdim=True)
        # N x C x L
        if self.elementwise_affine:
            x = self.weight * (x - mean) / torch.sqrt(var + self.eps) + self.bias
        else:
            x = (x - mean) / torch.sqrt(var + self.eps)
        return x


class CumulativeLayerNorm(nn.LayerNorm):
    '''
       Calculate Cumulative Layer Normalization
       dim: you want to norm dim
       elementwise_affine: learnable per-element affine parameters 
    '''

    def __init__(self, dim, elementwise_affine=True):
        super(CumulativeLayerNorm, self).__init__(
            dim, elementwise_affine=elementwise_affine)

    def forward(self, x):
        # x: N x C x L
        # N x L x C
        x = torch.transpose(x, 1, 2)
        # N x L x C == only channel norm
        x = super().forward(x)
        # N x C x L
        x = torch.transpose(x, 1, 2)
        return x


def select_norm(norm, dim):
    if norm == 'gln':
        return GlobalLayerNorm(dim, elementwise_affine=True)
    if norm == 'cln':
        return CumulativeLayerNorm(dim, elementwise_affine=True)
    else:
        return nn.BatchNorm1d(dim)


class Encoder(nn.Module):

    def __init__(self, in_channels, out_channels, kernel_size, stride):
        super(Encoder, self).__init__()
        self.sequential = nn.Sequential(
            Conv1D(in_channels, out_channels, kernel_size, stride=stride),
            Conv1D(out_channels, out_channels, kernel_size=3, stride=1, padding=1),
            nn.PReLU(),
            Conv1D(out_channels, out_channels, kernel_size=3, stride=1, padding=1),
            nn.PReLU(),
            Conv1D(out_channels, out_channels, kernel_size=3, stride=1, padding=1),
            nn.PReLU()
        )

    def forward(self, x):
        '''
           x: [B, T]
           out: [B, N, T]
        '''
        x = self.sequential(x)
        return x


class Decoder(nn.Module):
    '''
        Decoder
        This module can be seen as the gradient of Conv1d with respect to its input. 
        It is also known as a fractionally-strided convolution 
        or a deconvolution (although it is not an actual deconvolution operation).
    '''

    def __init__(self, N, kernel_size=16, stride=16 // 2):
        super(Decoder, self).__init__()
        self.sequential = nn.Sequential(
            nn.ConvTranspose1d(N, N, kernel_size=3, stride=1, padding=1),
            nn.PReLU(),
            nn.ConvTranspose1d(N, N, kernel_size=3, stride=1, padding=1),
            nn.PReLU(),
            nn.ConvTranspose1d(N, N, kernel_size=3, stride=1, padding=1),
            nn.PReLU(),
            nn.ConvTranspose1d(N, 1, kernel_size=kernel_size, stride=stride, bias=True)
        )

    def forward(self, x):
        """
        x: N x L or N x C x L
        """
        x = self.sequential(x)
        if torch.squeeze(x).dim() == 1:
            x = torch.squeeze(x, dim=1)
        else:
            x = torch.squeeze(x)

        return x


class Conv1D(nn.Conv1d):
    '''
       Applies a 1D convolution over an input signal composed of several input planes.
    '''

    def __init__(self, *args, **kwargs):
        super(Conv1D, self).__init__(*args, **kwargs)

    def forward(self, x, squeeze=False):
        # x: N x C x L
        if x.dim() not in [2, 3]:
            raise RuntimeError("{} accept 2/3D tensor as input".format(
                self.__name__))
        x = super().forward(x if x.dim() == 3 else torch.unsqueeze(x, 1))
        if squeeze:
            x = torch.squeeze(x)
        return x


class ConvTrans1D(nn.ConvTranspose1d):
    '''
       This module can be seen as the gradient of Conv1d with respect to its input. 
       It is also known as a fractionally-strided convolution 
       or a deconvolution (although it is not an actual deconvolution operation).
    '''

    def __init__(self, *args, **kwargs):
        super(ConvTrans1D, self).__init__(*args, **kwargs)

    def forward(self, x, squeeze=False):
        """
        x: N x L or N x C x L
        """
        if x.dim() not in [2, 3]:
            raise RuntimeError("{} accept 2/3D tensor as input".format(
                self.__name__))
        x = super().forward(x if x.dim() == 3 else torch.unsqueeze(x, 1))
        if squeeze:
            x = torch.squeeze(x)
        return x


class Conv1D_Block(nn.Module):
    '''
       Consider only residual links
    '''

    def __init__(self, in_channels=256, out_channels=512,
                 kernel_size=3, dilation=1, norm='gln', causal=False, skip_con='True'):
        super(Conv1D_Block, self).__init__()
        # conv 1 x 1
        self.conv1x1 = Conv1D(in_channels, out_channels, 1)
        self.PReLU_1 = nn.PReLU()
        self.norm_1 = select_norm(norm, out_channels)
        # not causal don't need to padding, causal need to pad+1 = kernel_size
        self.pad = (dilation * (kernel_size - 1)) // 2 if not causal else (
                dilation * (kernel_size - 1))
        # depthwise convolution
        self.dwconv = Conv1D(out_channels, out_channels, kernel_size,
                             groups=out_channels, padding=self.pad, dilation=dilation)
        self.PReLU_2 = nn.PReLU()
        self.norm_2 = select_norm(norm, out_channels)
        self.Sc_conv = nn.Conv1d(out_channels, in_channels, 1, bias=True)
        self.Output = nn.Conv1d(out_channels, in_channels, 1, bias=True)
        self.causal = causal
        self.skip_con = skip_con

    def forward(self, x):
        # x: N x C x L
        # N x O_C x L
        c = self.conv1x1(x)
        # N x O_C x L
        c = self.PReLU_1(c)
        c = self.norm_1(c)
        # causal: N x O_C x (L+pad)
        # noncausal: N x O_C x L
        c = self.dwconv(c)
        c = self.PReLU_2(c)
        c = self.norm_2(c)
        # N x O_C x L
        if self.causal:
            c = c[:, :, :-self.pad]
        if self.skip_con:
            Sc = self.Sc_conv(c)
            c = self.Output(c)
            return Sc, c + x
        c = self.Output(c)
        return x + c


class Separation(nn.Module):
    """
       R	Number of repeats
       X	Number of convolutional blocks in each repeat
       B	Number of channels in bottleneck and the residual paths’ 1 × 1-conv blocks
       H	Number of channels in convolutional blocks
       P	Kernel size in convolutional blocks
       norm The type of normalization(gln, cl, bn)
       causal  Two choice(causal or noncausal)
       skip_con Whether to use skip connection
    """

    def __init__(self, R, X, B, H, P, norm='gln', causal=False, skip_con=True):
        super(Separation, self).__init__()
        self.separation = nn.ModuleList([])
        for r in range(R):
            for x in range(X):
                self.separation.append(Conv1D_Block(
                    B, H, P, 2 ** x, norm, causal, skip_con))
        self.skip_con = skip_con

    def forward(self, x):
        '''
           x: [B, N, L]
           out: [B, N, L]
        '''
        if self.skip_con:
            skip_connection = 0
            for i in range(len(self.separation)):
                skip, out = self.separation[i](x)
                skip_connection = skip_connection + skip
                x = out
            return skip_connection
        else:
            for i in range(len(self.separation)):
                out = self.separation[i](x)
                x = out
            return x


class ConvTasNet(nn.Module):
    '''
       ConvTasNet module
       N	Number of ﬁlters in autoencoder
       L	Length of the ﬁlters (in samples)
       B	Number of channels in bottleneck and the residual paths’ 1 × 1-conv blocks
       Sc	Number of channels in skip-connection paths’ 1 × 1-conv blocks
       H	Number of channels in convolutional blocks
       P	Kernel size in convolutional blocks
       X	Number of convolutional blocks in each repeat
       R	Number of repeats
    '''

    def __init__(self,
                 N=512,
                 L=16,
                 B=128,
                 H=512,
                 P=3,
                 X=8,
                 R=3,
                 norm="gln",
                 num_spks=2,
                 activate="relu",
                 causal=False,
                 skip_con=False):
        super(ConvTasNet, self).__init__()
        # n x 1 x T => n x N x T
        self.encoder = Encoder(1, N, L, stride=L // 2)
        # n x N x T  Layer Normalization of Separation
        self.LayerN_S = select_norm('cln', N)
        # n x B x T  Conv 1 x 1 of  Separation
        self.BottleN_S = Conv1D(N, B, 1)
        # Separation block
        # n x B x T => n x B x T
        self.separation = Separation(R, X, B, H, P, norm=norm, causal=causal, skip_con=skip_con)
        # n x B x T => n x 2*N x T
        self.gen_masks = Conv1D(B, num_spks * N, 1)
        # n x N x T => n x 1 x L
        self.decoder = Decoder(N, L, stride=L // 2)
        # activation function
        active_f = {
            'relu': nn.ReLU(),
            'sigmoid': nn.Sigmoid(),
            'softmax': nn.Softmax(dim=0)
        }
        self.activation_type = activate
        self.activation = active_f[activate]
        self.num_spks = num_spks

    def forward(self, x):
        if x.dim() >= 3:
            raise RuntimeError(
                "{} accept 1/2D tensor as input, but got {:d}".format(
                    self.__name__, x.dim()))
        if x.dim() == 1:
            x = torch.unsqueeze(x, 0)
        # x: n x 1 x L => n x N x T
        w = self.encoder(x)
        # n x N x L => n x B x L
        e = self.LayerN_S(w)
        e = self.BottleN_S(e)
        # n x B x L => n x B x L
        e = self.separation(e)
        # n x B x L => n x num_spk*N x L
        m = self.gen_masks(e)
        # n x N x L x num_spks
        m = torch.chunk(m, chunks=self.num_spks, dim=1)
        # num_spks x n x N x L
        m = self.activation(torch.stack(m, dim=0))
        d = [w * m[i] for i in range(self.num_spks)]
        # decoder part num_spks x n x L
        s = [self.decoder(d[i]) for i in range(self.num_spks)]
        return s


def test_convtasnet():
    x = torch.randn(1, 32000)
    nnet = ConvTasNet()

    s = nnet(x)

    summary(nnet, input_size=(1, 32000))
    # print(s.shape)


if __name__ == "__main__":
    test_convtasnet()



import numpy as np
import torch
import numpy.random as random
from .utils import load_file
import os
from pathlib import Path


class AddNoise(torch.nn.Module):

    def __init__(self, noise_file_paths: str = None,
                 key="acc[2]",
                 add_prob: float = 0.8):
        """
        :param noise_file_paths: file path for noise
        """
        super().__init__()
        self.noise_list = []
        self.key = key
        self.add_prob = add_prob
        noise_file_paths = Path(noise_file_paths)
        for path in sorted(noise_file_paths.rglob('*')):
            full_path_str = str(path)  # Get the full path string
            if path.is_file() and "clean" not in full_path_str:
                # print(full_path_str)
                self.noise_list.append(full_path_str)

    def forward(self, x):
        if random.rand() > self.add_prob:
            # print("No noise added")
            return x
        index = random.randint(0, len(self.noise_list))
        noise_file_path = self.noise_list[index]
        noise = load_file(noise_file_path, self.key)
        x = x + noise[0].reshape(1, -1)
        return x



from torchsense.models.lit_model import *
import lightning as L
import os
from lightning.pytorch.callbacks import LearningRateMonitor, ModelCheckpoint
from lightning.pytorch.loggers import CSVLogger
import torch
from lightning.pytorch.tuner import Tuner
import time
from pathlib import Path
L.seed_everything(42)

torch.set_float32_matmul_precision("medium")


class Trainer:
    def __init__(self, model, save_name = "base",loss=None, task="r", precision="32", max_epochs=5,
                 lr=0.001, logger = None,*args):
        if task == "r":
            self.model = LitRegressModel(model, lr=lr, loss_fn=loss)
        elif task == "c":
            self.model = LitClassModel(model, lr=lr)
        elif task == "m":
            self.model = LitMultimodalModel(model, lr=lr)
        elif task == "two" or task == "t":
            self.model = LitTwoStageModel(model, lr=lr)
        else:
            raise ValueError("task must be either 'r' or 'c'")
        CHECKPOINT_PATH = Path("outputs-lightnings", save_name,time.strftime("%Y%m%d/%H%M%S"))

        print("model saved at :",CHECKPOINT_PATH/"ckpt")
        os.makedirs(CHECKPOINT_PATH/"ckpt", exist_ok=True)
        if logger is None:
            logger = CSVLogger(CHECKPOINT_PATH,name=None)
        self.max_epochs = max_epochs
        self.precision = precision
        self.trainer = L.Trainer(
            precision=self.precision,
            accelerator="auto",
            max_epochs=self.max_epochs,
            logger=logger,
            callbacks=[
                ModelCheckpoint(dirpath=CHECKPOINT_PATH/"ckpt", filename='{epoch}-{val_loss_epoch:.2f}',
                                save_weights_only=True, mode="min", monitor="val_loss_epoch",
                                enable_version_counter=False,
                                save_last=True, save_top_k=3),
                # Save the best checkpoint based on the maximum val_acc recorded. Saves only weights and not optimizer

            ], *args,
        )

    def fit(self, train_loader, val_loader):

        self.trainer.fit(self.model, train_loader, val_loader)


    def lr_find(self, train_loader, val_loader):

        tuner = Tuner(self.trainer)

        # finds learning rate automatically
        # sets hparams.lr or hparams.learning_rate to that learning rate
        lr_finder = tuner.lr_find(self.model, train_loader, val_loader)
        print(lr_finder.suggestion())



from torchsense.trainer import Trainer
from torchsense.datasets.custom import SensorFolder
from torch.utils.data import DataLoader
from torchsense.models import Generator, UNet
from torchsense import transforms as T
from torchaudio.transforms import Spectrogram,GriffinLim
from torchsense.utils import load_from_ckpt
import torchaudio


def train():
    # data part
    data_path = "data1"
    transform1 = T.Compose([
        T.ToTensor(),
        T.Normalize(-1, 1),
        Spectrogram(n_fft=512, hop_length=160, win_length=256, power=1),
    ])
    transform2 = T.Compose([
        T.ToTensor(),
        T.Interpolate(5000),
        Spectrogram(n_fft=100, hop_length=10, win_length=100, power=1),
    ])
    target_transform = T.Compose([
        T.ToTensor(),
        T.Normalize(-1, 1),
        Spectrogram(n_fft=512, hop_length=160, win_length=256, power=1),
    ])
    stage_transform = T.Compose([
        # T.Normalize(-1, 1),
        GriffinLim(n_fft=512, hop_length=160, win_length=256, power=1),
    ])
    stage1_model = Generator()
    pre_model = load_from_ckpt(stage1_model, "epoch=19-val_loss=0.80.ckpt")
    data = SensorFolder(root=data_path,
                        params=(["acc[2]", "mix_mic"], ["mic"]),
                        max_samples=1000,
                        # pre_model=pre_model,
                        # stage_transform=stage_transform,
                        transform=[transform2, transform1],
                        target_transform=target_transform)
    train_set, test_set = data.train_test_split(0.5)
    train_loader = DataLoader(train_set, batch_size=1, shuffle=True,
                              drop_last=True, num_workers=0, pin_memory=True)
    val_loader = DataLoader(test_set, batch_size=1, shuffle=False,
                            drop_last=True, num_workers=0)

    # model part

    model = Generator()
    # training part
    trainer = Trainer(model,  max_epochs=5, task="r", )
    trainer.fit(train_loader, val_loader)


if __name__ == "__main__":
    train()



from .utils import load_file
from .folder import DatasetFolder
from typing import Any, Callable, cast, Dict, List, Optional, Tuple, Union
from torchsense.transforms import has_batch_dimension,add_batch_dimension,remove_batch_dimension, Compose
from typing import List, Callable, Any

IMG_EXTENSIONS = (".mat", ".jpeg", ".npz")


def default_loader(path: str, params: list) -> Any:
    return load_file(path, params)


class SensorFolder(DatasetFolder):
    """A generic data1 loader where the images are arranged in this way by default: ::

        root/dog/xxx.png
        root/dog/xxy.png
        root/dog/[...]/xxz.png

        root/cat/123.png
        root/cat/nsdf3.png
        root/cat/[...]/asd932_.png

    This class inherits from :class:`~torchvision.datasets.DatasetFolder` so
    the same methods can be overridden to customize the dataset.

    Args:
        root (str or ``pathlib.Path``): Root directory path.
        transform (callable, optional): A function/transform that takes in a PIL image
            and returns a transformed version. E.g, ``transforms.RandomCrop``
        target_transform (callable, optional): A function/transform that takes in the
            target and transforms it.
        loader (callable, optional): A function to load an image given its path.
        is_valid_file (callable, optional): A function that takes path of an Image file
            and check if the file is a valid file (used to check of corrupt files)
        allow_empty(bool, optional): If True, empty folders are considered to be valid classes.
            An error is raised on empty folders if False (default).

     Attributes:
        classes (list): List of the class names sorted alphabetically.
        class_to_idx (dict): Dict with items (class_name, class_index).
        imgs (list): List of (image path, class_index) tuples
    """

    def __init__(
            self,
            root: str,
            params: Tuple[List[str], Optional[List[str]]],
            pre_model=None,
            stage_transform: Optional[Callable] = None,
            max_samples: Optional[int] = None,
            transform: Union[Optional[Callable], List[Callable]] = None,
            target_transform: Optional[Callable] = None,
            loader: Callable[[str, Any], Any] = default_loader,
            is_valid_file: Optional[Callable[[str], bool]] = None,
            allow_empty: bool = False,
    ):
        super().__init__(
            root,
            params,
            loader,
            IMG_EXTENSIONS if is_valid_file is None else None,
            transform=transform,
            target_transform=target_transform,
            is_valid_file=is_valid_file,
            allow_empty=allow_empty,
        )
        if max_samples is not None:
            max_samples = min(max_samples, len(self.samples))
            self.samples = self.samples[:max_samples]
        match pre_model:
            case None:
                self.pre_model = None
            case _:
                self.pre_model = pre_model
                self.stage_transform = stage_transform if stage_transform is not None else None

    def __getitem__(self, index: int) -> Tuple[Any, Any, Any]:
        """
        Args:
            index (int): Index

        Returns:
            tuple: (sample, target) where target is class_index of the target class.
        """
        path, labels = self.samples[index]
        sample, target = self.loader(path, self.params)
        if len(sample) == 1:
            sample = sample[0]
        if len(target) == 1:
            target = target[0]

        if self.pre_model is not None:
            if has_batch_dimension(sample):
                sample = add_batch_dimension(sample)
            stage_out = self.pre_model(tuple(sample))
            stage_out = remove_batch_dimension(stage_out)
            if self.stage_transform is not None:
                sample = self.stage_transform(stage_out)

        if self.transform is not None:
            if isinstance(self.transform, Compose):
                sample = self.transform(sample)
            else:
                for i in range(len(self.transform)):
                    sample[i] = self.transform[i](sample[i])

        if self.target_transform is not None:
            target = self.target_transform(target)

        return sample, target, labels

    def apply_transform(self, sample: Any) -> Any:
        if self.transform is not None:
            match self.transform:
                case Compose():
                    sample = self.transform(sample)
                case list() as transform_list if all(callable(t) for t in transform_list):
                    for i in range(len(transform_list)):
                        sample[i] = transform_list[i](sample[i])
                case _:
                    raise TypeError("transform should be either a Compose or a list of callables.")
        return sample



# from models import BaseVAE
import torch
from torch import nn
import mobilenet


class DepthSeparableConv1d(nn.Module):

    def __init__(self, input_channels, out_channels, kernel_size, **kwargs):
        super().__init__()
        self.Depthwise = nn.Sequential(
            nn.Conv1d(
                input_channels, input_channels,
                kernel_size, groups=input_channels, **kwargs),
            nn.BatchNorm1d(input_channels),
            nn.LeakyReLU(inplace=True)
        )

        self.Pointwise = nn.Sequential(
            nn.Conv1d(input_channels, out_channels, 1),
            nn.BatchNorm1d(out_channels),
            nn.LeakyReLU(inplace=True)
        )

    def forward(self, x):
        x = self.Depthwise(x)
        x = self.Pointwise(x)
        return x


class DepthSeparableConvTranspose1d(nn.Module):

    def __init__(self, input_channels, out_channels, kernel_size, **kwargs):
        super().__init__()
        self.Depthwise = nn.Sequential(
            nn.ConvTranspose1d(input_channels, input_channels,
                               kernel_size, groups=input_channels, **kwargs),
            nn.BatchNorm1d(input_channels),
            nn.LeakyReLU(inplace=True)
        )

        self.Pointwise = nn.Sequential(
            nn.Conv1d(input_channels, out_channels, 1),
            nn.BatchNorm1d(out_channels),
            nn.LeakyReLU(inplace=True)
        )

    def forward(self, x):
        x = self.Depthwise(x)
        # print("dw:", x.size())
        x = self.Pointwise(x)
        # print("pw:", x.size())
        # print("--------------")
        return x


class BasicConv1d(nn.Module):

    def __init__(self, input_channels, out_channels, kernel_size, **kwargs):
        super().__init__()
        self.conv = nn.Conv1d(
            input_channels, out_channels, kernel_size, **kwargs)
        self.bn = nn.BatchNorm1d(out_channels)
        self.relu = nn.LeakyReLU()

    def forward(self, x):
        x = self.conv(x)
        x = self.bn(x)
        x = self.relu(x)

        return x


class LinearBottleNeck(nn.Module):

    def __init__(self, in_channels, out_channels, stride, t=6):
        super(LinearBottleNeck, self).__init__()

        self.residual = nn.Sequential(
            # pw
            nn.Conv1d(in_channels, in_channels * t, 1, bias=False),
            nn.BatchNorm1d(in_channels * t),
            nn.ReLU6(inplace=True),
            # dw
            nn.Conv1d(in_channels * t, in_channels * t,
                      3, stride=stride, padding=1, groups=in_channels * t, bias=False),
            nn.BatchNorm1d(in_channels * t),
            nn.ReLU6(inplace=True),
            # pw
            nn.Conv1d(in_channels * t, out_channels, 1, bias=False),
            nn.BatchNorm1d(out_channels)
        )

        self.stride = stride
        self.in_channels = in_channels
        self.out_channels = out_channels

    def forward(self, x):
        residual = self.residual(x)
        if self.stride == 1 and self.in_channels == self.out_channels:
            residual += x

        return residual


class MobileNetV1(nn.Module):

    def __init__(self, in_channels: int) -> None:
        super().__init__()
        modules = [
            BasicConv1d(in_channels, out_channels=32,
                        kernel_size=3, stride=2, padding=1),
            DepthSeparableConv1d(32, out_channels=64,
                                 kernel_size=3, stride=2, padding=1, bias=False),
            DepthSeparableConv1d(64, out_channels=128,
                                 kernel_size=3, stride=1, padding=1, bias=False),
            DepthSeparableConv1d(128, out_channels=256,
                                 kernel_size=3, stride=2, padding=1, bias=False),
            DepthSeparableConv1d(256, out_channels=512,
                                 kernel_size=3, stride=2, padding=1, bias=False),
            DepthSeparableConv1d(512, out_channels=512,
                                 kernel_size=3, stride=2, padding=1, bias=False)
        ]

        # Build Encoder 通过Sequential将网络层和激活函数结合起来，输出激活后的五层网络节点
        self.encoder = nn.Sequential(*modules)  # 将输入i加载到网络上
        self.low_level_features = self.encoder[0:1]
        self.middle_level_feature = self.encoder[1]
        self.high_level_features = self.encoder[2:]

    def forward(self, x):
        low_level_feat = self.low_level_features(x)
        # print("low after backbone", x.shape)
        middle_level_feat = self.middle_level_feature(low_level_feat)
        x = self.high_level_features(middle_level_feat)
        # print("x after backbone:", x.size())
        return low_level_feat, middle_level_feat, x


def build_backbone(backbone, in_channels):

    if backbone == 'mobilenet':
        return mobilenet.MobileNetV1(in_channels)
    else:
        raise


import numpy as np
import torch
import torch.nn as nn
from torch.autograd import Variable
from typing import List
from utils import show_parameter


class ConvLSTMCell(nn.Module):
    def __init__(self, input_channels, hidden_channels):
        super(ConvLSTMCell, self).__init__()

        # assert hidden_channels % 2 == 0

        self.input_channels = input_channels
        self.hidden_channels = hidden_channels
        self.kernel_size = 3
        self.num_features = 4

        self.padding = int((self.kernel_size - 1) / 2)
        # 遗忘门的 neural network layer 构成
        self.Wfx = nn.Conv1d(self.input_channels, self.hidden_channels, self.kernel_size, 1, self.padding, bias=True)
        self.Wfh = nn.Conv1d(self.hidden_channels, self.hidden_channels, self.kernel_size, 1, self.padding, bias=False)

        # 输入门的 neural network layer 构成
        self.Wix = nn.Conv1d(self.input_channels, self.hidden_channels, self.kernel_size, 1, self.padding, bias=True)
        self.Wih = nn.Conv1d(self.hidden_channels, self.hidden_channels, self.kernel_size, 1, self.padding, bias=False)

        # 激活单元的 neural network layer 构成
        self.Wcx = nn.Conv1d(self.input_channels, self.hidden_channels, self.kernel_size, 1, self.padding, bias=True)
        self.Wch = nn.Conv1d(self.hidden_channels, self.hidden_channels, self.kernel_size, 1, self.padding, bias=False)

        # 输出门的 neural network layer 构成
        self.Wox = nn.Conv1d(self.input_channels, self.hidden_channels, self.kernel_size, 1, self.padding, bias=True)
        self.Woh = nn.Conv1d(self.hidden_channels, self.hidden_channels, self.kernel_size, 1, self.padding, bias=False)

        self.Wic = None
        self.Wfc = None
        self.Woc = None

    def forward(self, x, h, c):
        # 遗忘门
        cf = torch.sigmoid(self.Wfx(x) + self.Wfh(h) + c * self.Wfc)
        # 更新门
        ci = torch.sigmoid(self.Wix(x) + self.Wih(h) + c * self.Wic)
        # 前一时刻*forget+现一时刻*input
        cc = cf * c + ci * torch.tanh(self.Wcx(x) + self.Wch(h))
        # 输出门
        co = torch.sigmoid(self.Wox(x) + self.Woh(h) + cc * self.Woc)
        ch = co * torch.tanh(cc)
        return ch, cc

    def init_hidden(self, batch_size, hidden, shape):
        if self.Wic is None:
            self.Wic = nn.Parameter(torch.zeros(1, hidden, shape[0])).cuda()
            self.Wfc = nn.Parameter(torch.zeros(1, hidden, shape[0])).cuda()
            self.Woc = nn.Parameter(torch.zeros(1, hidden, shape[0])).cuda()
        # else:
        #  assert shape[0] == self.Wic.size()[2], 'Input Length Mismatched!'
        return (torch.zeros((batch_size, hidden, shape[0])).cuda(),
                torch.zeros((batch_size, hidden, shape[0])).cuda())


class ConvLSTM(nn.Module):
    # input_channels corresponds to the first input feature map
    # hidden state is 2 list of succeeding lstm layers.
    def __init__(self, batch_size, input_channels, effective_step):
        super(ConvLSTM, self).__init__()
        hidden_channels = [32]
        self.input_channels = [input_channels] + hidden_channels
        # print("input:", [input_channels] + hidden_channels)
        self.hidden_channels = hidden_channels
        self.num_layers = len(hidden_channels)
        self.batch_size = batch_size
        self.step = 10
        self.effective_step = effective_step
        self._all_layers = []
        # num_layers: 纵向有多少个时间单元
        for i in range(self.num_layers):
            name = 'cell{}'.format(i)
            cell = ConvLSTMCell(self.input_channels[i], self.hidden_channels[i])
            # 将name的值变成cell
            setattr(self, name, cell)
            self._all_layers.append(cell)
        self.final_layer = nn.Sequential(
            nn.Conv1d(hidden_channels[-1], out_channels=1,
                      kernel_size=3, padding=1),
            nn.Dropout(0.5),
            nn.Linear(1 * 3840, 64),
            nn.Dropout(0.5),
            nn.Linear(64, 2)
        )

    def get_reconstruction_loss(self, *args, labels) -> dict:
        recons = args[0]
        recons_loss_ = nn.SmoothL1Loss()
        # recons_loss = recons_loss_(recons.view(4, -1), gt.view(4, -1))  # reshape成20行，列数由行数决定
        recons_loss = recons_loss_(recons, labels)
        loss = recons_loss
        return loss

    def forward(self, input):
        internal_state = []
        # outputs = torch.zeros(self.batch_size, self.hidden_channels[-1], input.shape[2]).cuda()
        outputs = []
        # step:横向有多少个时间步
        for step in range(self.step):
            # print(step)
            x = input
            for i in range(self.num_layers):
                # all cells are initialized in the first step
                name = 'cell{}'.format(i)
                if step == 0:
                    bsize, _, length = x.size()
                    (h, c) = getattr(self, name).init_hidden(batch_size=bsize, hidden=self.hidden_channels[i],
                                                             shape=[length])
                    internal_state.append((h, c))  # 纵向传播
                # do forward
                (h, c) = internal_state[i]
                x, new_c = getattr(self, name)(x, h, c)
                internal_state[i] = (x, new_c)
            # only record effective steps
            if step < self.effective_step:
                # print(x.shape())
                outputs.append(x)
            # outputs = np.array(outputs)
            # outputs.to(torch.device("cuda"))
        outputs = self.final_layer(outputs[-1])
        return outputs, (x, new_c)


if __name__ == '__main__':
    # gradient check
    batch_size = 10
    model = ConvLSTM(batch_size, input_channels=1, effective_step=2).cuda()
    loss_fn = torch.nn.MSELoss()
    show_parameter(model)
    input = Variable(torch.randn(batch_size, 1, 3840)).cuda()
    target = Variable(torch.randn(batch_size, 1, 2)).double().cuda()

    output = model(input)
    output = output[0][0].double()
    res = torch.autograd.gradcheck(loss_fn, (output, target), eps=1e-6, raise_exception=True)
    print(model)
    print(output.size())
